/// ===========================================================================
/// C= Parallel Programming Language Runtime
/// ===========================================================================
/// Copyright HOOPOE ProGroup, 2010-2011. All rights reserved
/// ---------------------------------------------------------------------------
/// VERSION: 0.3
/// ---------------------------------------------------------------------------
/// FILE: cstripes.c
/// ---------------------------------------------------------------------------

#include "internal.h"

/// ---------------------------------------------------------------------------
/// Global Data
/// ---------------------------------------------------------------------------

/// initialization/destruction
static pthread_key_t cpx_global_tls_key = -1;
static volatile int cpx_global_cleanup = 0;

/// hardware info
static int cpx_global_cpuno = 0;

/// thread pooling and task management
static volatile int cpx_global_task_no = 1;
static volatile int cpx_global_task_lock = 0;
static volatile int cpx_global_thread_no = 0;
static volatile int cpx_global_free_thread_no = 0;

static sem_t cpx_global_free_semaphore;
static sem_t cpx_global_task_semaphore;

static cpx_datatype_region_context* volatile cpx_global_task_list_head = 0;

/// object-free synchronization
static cpx_datatype_sync_queue cpx_global_sync_queue[CPX_CONST_MAX_SYNC_HASH];

/// ---------------------------------------------------------------------------
/// Initialization / Destruction
/// ---------------------------------------------------------------------------

void __attribute__((constructor)) init_module(void)
{
    /// initialize global data and create thread pools
    cpx_intrinsic_global_init();
    /// allocate TLS structures for the current thread
    cpx_intrinsic_init_tls();
}

void __attribute__((destructor)) cleanup_module(void)
{
    /// destroy global structures and thread pools
    cpx_intrinsic_global_cleanup(0);
}

static int cpx_intrinsic_get_cpuno()
{
    int res = sysconf(_SC_NPROCESSORS_ONLN);

    res = res < 1 ? 1 : res;

    return res;
}

int cpx_intrinsic_global_init()
{
    int i;
    int res = 0;

    /// initialize global objects
    if(cpx_global_tls_key == -1)
    {
        /// allocate TLS index
        if(pthread_key_create(&cpx_global_tls_key, 
                              (void (*)(void*))cpx_intrinsic_free_tls))
        {
            /// serves as a general initialization flag
            cpx_global_tls_key = -1;

            return -1;
        }
        /// initialize global objects
        res += sem_init(&cpx_global_free_semaphore, 0, 0);
        res += sem_init(&cpx_global_task_semaphore, 0, 0);

        if(res)
        {
            /// uninitialize
            cpx_global_tls_key = -1;
            return -2;
        }
        /// obtain the number of system processors
        cpx_global_cpuno = cpx_intrinsic_get_cpuno();
        /// create a thread pool
        for(i = 1; i < cpx_global_cpuno; i++)
        {
            if(cpx_intrinsic_create_thread())
            {
                break;
            }
        }
        cpx_global_cpuno = i;
        /// initialize synchronization hash
        for(i = 0; i < CPX_CONST_MAX_SYNC_HASH; i++)
        {
            cpx_global_sync_queue[i].win_count = 0;
            cpx_global_sync_queue[i].winner = 0;
            cpx_global_sync_queue[i].lock = 0;
            cpx_global_sync_queue[i].lcnt = 0;
            cpx_global_sync_queue[i].head = 0;
        }
    }
    return 0;
}

void cpx_intrinsic_init_tls()
{
    cpx_datatype_tls_context* tlsctx;

    pthread_mutexattr_t attr;

    /// initialize global objects
    if(cpx_global_tls_key == -1)
    {
        if(cpx_intrinsic_global_init())
        {
            return;
        }
    }
    /// allocate and initialize thread-specific objects
    if(!(tlsctx = malloc(sizeof(cpx_datatype_tls_context))))
    {
        return;
    }
    memset(tlsctx, 0, sizeof(cpx_datatype_tls_context));

    pthread_mutexattr_init(&attr); 
    pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE); 
    pthread_mutex_init(&tlsctx->mutex, &attr);

    if(sem_init(&tlsctx->wakesig, 0, 0))
    {
        free(tlsctx);
        return;
    }

    pthread_setspecific(cpx_global_tls_key, tlsctx);
}

/// called implicitly upon TLS destruction
void cpx_intrinsic_free_tls(cpx_datatype_tls_context* tlsctx)
{
    if(tlsctx)
    {
        if(!tlsctx->rgnctx)
        {
            pthread_setspecific(cpx_global_tls_key, 0);
            pthread_mutex_destroy(&tlsctx->mutex);
            sem_destroy(&tlsctx->wakesig);
            free(tlsctx);
        }
    }
}

/// to be called explicitly before exiting the main thread
void cpx_intrinsic_global_cleanup(int wait)
{
    int i;
    struct timespec req = {0, 100000000};
    struct timespec rem;

    /// explicit destruction of the current thread's TLS
    cpx_intrinsic_free_tls(cpx_intrinsic_get_tlsctx());

    /// destroy threads and global objects
    cpx_global_cleanup = cpx_global_thread_no;

    for(i = 0; i < cpx_global_thread_no; i++)
    {
        sem_post(&cpx_global_free_semaphore);
        sem_post(&cpx_global_task_semaphore);
    }
    nanosleep(&req, &rem);

    if(wait)
    {
        while(cpx_global_cleanup)
        {
            nanosleep(&req, &rem);
        }
    }
    sem_destroy(&cpx_global_free_semaphore);
    sem_destroy(&cpx_global_task_semaphore);

    pthread_key_delete(cpx_global_tls_key);
    cpx_global_tls_key = -1;
}

/// ---------------------------------------------------------------------------
/// Common
/// ---------------------------------------------------------------------------

cpx_datatype_tls_context* cpx_intrinsic_get_tlsctx()
{
    cpx_datatype_tls_context* tlsctx;

    if(cpx_global_tls_key == -1 ||
        !(tlsctx = pthread_getspecific(cpx_global_tls_key)))
    {
        cpx_intrinsic_init_tls();

        if(cpx_global_tls_key == -1 || 
            !(tlsctx = pthread_getspecific(cpx_global_tls_key)))
        {
            /// TODO: Throw an exception in case of error?
            return 0;
        }
    }
    return tlsctx;
}

/// ---------------------------------------------------------------------------
/// Thread Pool
/// ---------------------------------------------------------------------------

static void* cpx_intrinsic_thread(void* arg)
{
    /// initialize
    cpx_datatype_tls_context* tlsctx = cpx_intrinsic_get_tlsctx();

    if(!tlsctx)
    {
        (void)__sync_fetch_and_sub(&cpx_global_thread_no, 1);
        return (void*)-1;
    }
    for(;;)
    {
        /// check for cleanup signal
        if(cpx_global_cleanup)
        {
            cpx_intrinsic_free_tls(tlsctx);
            (void)__sync_fetch_and_sub(&cpx_global_cleanup, 1);
            (void)__sync_fetch_and_sub(&cpx_global_thread_no, 1);
            break;
        }
        /// arbitrate for task pool
        if(__sync_add_and_fetch(&cpx_global_task_no, 1) > cpx_global_cpuno)
        {
            (void)__sync_fetch_and_sub(&cpx_global_task_no, 1);
            (void)__sync_fetch_and_add(&cpx_global_free_thread_no, 1);
            sem_wait(&cpx_global_free_semaphore);
        }
        else
        {
            sem_wait(&cpx_global_task_semaphore);

            /// process tasks
            cpx_exported_pick_job();

            (void)__sync_fetch_and_sub(&cpx_global_task_no, 1);
        }
    }
    return 0;
}

/// creates one more thread and returns 0 in case of success
int cpx_intrinsic_create_thread()
{
    pthread_t thread;

    if(pthread_create(&thread, 0, cpx_intrinsic_thread, 0))
    {
        return -1;
    }
    pthread_detach(thread);

    (void)__sync_add_and_fetch(&cpx_global_thread_no, 1);

    return 0;
}

/// ---------------------------------------------------------------------------
/// Parallel Operator Support Functions
/// ---------------------------------------------------------------------------

static void cpx_intrinsic_lock_task_list()
{
    while(__sync_val_compare_and_swap(&cpx_global_task_lock, 0, 1));
}

static void cpx_intrinsic_unlock_task_list()
{
    cpx_global_task_lock = 0;
}

/// initializes and queues a task set
extern void cpx_exported_parallel(int x, cpx_datatype_region_context* rgnctx)
{
    int i;
    cpx_datatype_tls_context* tlsctx = cpx_intrinsic_get_tlsctx();

    if(!tlsctx)
    {
        /// create a single-task region
        x = 1;
    }
    /// decipher x
    x = x > 0 ? x : cpx_global_cpuno;

    /// initialize region context
    rgnctx->version = 0;
    rgnctx->type = cpx_const_task;
    rgnctx->pix = 0;
    rgnctx->pno = x;
    rgnctx->busy = x - 1;
    rgnctx->lcount = x - 1;
    rgnctx->bcount = x;
    rgnctx->head = tlsctx;
    rgnctx->idle = 0;

    rgnctx->head_rgn = rgnctx;

    rgnctx->curr = tlsctx;
    rgnctx->prev_nest = 0;
    rgnctx->prev_task = 0;

    /// queue tasks
    if(tlsctx)
    {
        rgnctx->prev_nest = tlsctx->rgnctx;
        tlsctx->rgnctx = rgnctx;
    }
    if(rgnctx->busy)
    {
        cpx_intrinsic_lock_task_list();

        rgnctx->prev_task = cpx_global_task_list_head;
        cpx_global_task_list_head = rgnctx;

        cpx_intrinsic_unlock_task_list();

        for(i = 0; i < x - 1; i++)
        {
            sem_post(&cpx_global_task_semaphore);
        }
    }
}

/// picks up and executes a task, returns 0 when no task is left
int cpx_exported_pick_job()
{
    cpx_datatype_region_context rgnctx;
    jmp_buf idle;
    cpx_datatype_tls_context* tlsctx = cpx_intrinsic_get_tlsctx();

    /// fetch a task {
    cpx_intrinsic_lock_task_list();

    if(!cpx_global_task_list_head)
    {
        cpx_intrinsic_unlock_task_list();
        return 0;
    }
    /// make a local copy of the parallel region context
    memcpy(&rgnctx, cpx_global_task_list_head, 
           sizeof(cpx_datatype_region_context));

    /// adjust the local context
    rgnctx.idle = &idle;
    rgnctx.type = cpx_const_task;
    rgnctx.pix = rgnctx.pno - rgnctx.lcount;

    /// adjust the original context and the task list
    if(!--cpx_global_task_list_head->lcount)
    {
        ///cpx_global_task_list_head->free = 1;
        cpx_global_task_list_head = rgnctx.prev_task;
    }
    cpx_intrinsic_unlock_task_list();
    /// }

    /// execute the task
    if(!setjmp(idle))
    {
        rgnctx.curr = tlsctx;
        rgnctx.prev_nest = tlsctx->rgnctx;

        tlsctx->rgnctx = &rgnctx;
        tlsctx->ctxset = 1;

        cpx_intrinsic_swap_exectx(&rgnctx.exectx);
    }
    /// idle returns here

    return cpx_global_task_list_head ? 1 : 0;
}

/// picks up and executes a task, returns 0 when no task is left
/// the task should not be on the upper direct line of nesting
int cpx_exported_pick_nested_job(cpx_datatype_region_context* currctx)
{
    cpx_datatype_region_context rgnctx;
    jmp_buf idle;
    cpx_datatype_tls_context* tlsctx = cpx_intrinsic_get_tlsctx();

    /// fetch a task {
    cpx_intrinsic_lock_task_list();

    if(!cpx_global_task_list_head)
    {
        cpx_intrinsic_unlock_task_list();
        return 0;
    }

    /// search the nesting line to detect if it is safe to pick a task
    if(!currctx->lcount && currctx->prev_nest)
    {
        for(;;)
        {
            currctx = currctx->prev_nest;

            if(!currctx)
            {
                cpx_intrinsic_unlock_task_list();
                return 0;
            }
            currctx = currctx->head_rgn;

            if(!currctx->lcount)
            {
                continue;
            }
            if(cpx_global_task_list_head != currctx)
            {
                break;
            }
            else
            {
                cpx_intrinsic_unlock_task_list();
                return 0;
            }
        }
    }
    ///
    
    /// make a local copy of the parallel region context
    memcpy(&rgnctx, cpx_global_task_list_head, 
           sizeof(cpx_datatype_region_context));

    /// adjust the local context
    rgnctx.idle = &idle;
    rgnctx.type = cpx_const_task;
    rgnctx.pix = rgnctx.pno - rgnctx.lcount;

    /// adjust the original context and the task list
    if(!--cpx_global_task_list_head->lcount)
    {
        ///cpx_global_task_list_head->free = 1;
        cpx_global_task_list_head = rgnctx.prev_task;
    }
    cpx_intrinsic_unlock_task_list();
    /// }

    /// execute the task
    if(!setjmp(idle))
    {
        rgnctx.curr = tlsctx;
        rgnctx.prev_nest = tlsctx->rgnctx;

        tlsctx->rgnctx = &rgnctx;
        tlsctx->ctxset = 1;

        cpx_intrinsic_swap_exectx(&rgnctx.exectx);
    }
    /// idle returns here

    return cpx_global_task_list_head ? 1 : 0;
}

void cpx_intrinsic_swap_exectx(cpx_datatype_execution_context* exectx)
{
    size_t len;
    size_t bp, sp, diff;

    /// new register context
    ucontext_t* dst = &exectx->regs;

    /// adjust registers
    #define FIX_REG(x) ((x) = (bp > (x) && (x) >= sp) ? (x) + diff : (x))

    #if defined(__i386__) && defined(__linux__)

    /// allocate new stack
    char* frame = alloca(exectx->stk_frame - 
                         dst->uc_mcontext.gregs[REG_ESP] + 512);

    /// align the new stack
    frame = (char*)((((size_t)frame + 256) & -256) | 
                    (dst->uc_mcontext.gregs[REG_ESP] & 0xff));

    /// copy stack frames
    len = (size_t)exectx->stk_frame - dst->uc_mcontext.gregs[REG_ESP];

    memcpy(frame, (char*)dst->uc_mcontext.gregs[REG_ESP], len);

    /// source base pointer
    bp = exectx->stk_frame;
    /// source stack pointer
    sp = dst->uc_mcontext.gregs[REG_ESP];
    /// the distance between the destination and source stacks
    diff = (size_t)frame - sp;

    FIX_REG(dst->uc_mcontext.gregs[REG_EBP]);
    FIX_REG(dst->uc_mcontext.gregs[REG_EDI]);
    FIX_REG(dst->uc_mcontext.gregs[REG_ESI]);

    FIX_REG(dst->uc_mcontext.gregs[REG_EBX]);
    FIX_REG(dst->uc_mcontext.gregs[REG_ECX]);
    FIX_REG(dst->uc_mcontext.gregs[REG_EDX]);

    dst->uc_mcontext.gregs[REG_ESP] = (size_t)frame;

    #elif defined(__x86_64__) && defined(__linux__)

    /// allocate new stack
    char* frame = alloca(exectx->stk_frame - 
                         dst->uc_mcontext.gregs[REG_RSP] + 512);

    /// align the new stack
    frame = (char*)((((size_t)frame + 256) & -256) | 
                    (dst->uc_mcontext.gregs[REG_RSP] & 0xff));

    /// copy stack frames
    len = (size_t)exectx->stk_frame - dst->uc_mcontext.gregs[REG_RSP];

    memcpy(frame, (char*)dst->uc_mcontext.gregs[REG_RSP], len);

    /// source base pointer
    bp = exectx->stk_frame;
    /// source stack pointer
    sp = dst->uc_mcontext.gregs[REG_RSP];
    /// the distance between the destination and source stacks
    diff = (size_t)frame - sp;

    FIX_REG(dst->uc_mcontext.gregs[REG_RBP]);
    FIX_REG(dst->uc_mcontext.gregs[REG_RDI]);
    FIX_REG(dst->uc_mcontext.gregs[REG_RSI]);

    FIX_REG(dst->uc_mcontext.gregs[REG_RBX]);
    FIX_REG(dst->uc_mcontext.gregs[REG_RCX]);
    FIX_REG(dst->uc_mcontext.gregs[REG_RDX]);

    FIX_REG(dst->uc_mcontext.gregs[REG_R8]);
    FIX_REG(dst->uc_mcontext.gregs[REG_R9]);
    FIX_REG(dst->uc_mcontext.gregs[REG_R10]);
    FIX_REG(dst->uc_mcontext.gregs[REG_R11]);
    FIX_REG(dst->uc_mcontext.gregs[REG_R12]);
    FIX_REG(dst->uc_mcontext.gregs[REG_R13]);
    FIX_REG(dst->uc_mcontext.gregs[REG_R14]);
    FIX_REG(dst->uc_mcontext.gregs[REG_R15]);

    dst->uc_mcontext.gregs[REG_RSP] = (size_t)frame;

    #elif defined(__x86_64__) && defined(__FreeBSD__)

    /// allocate new stack
    char* frame = alloca(exectx->stk_frame - 
                         dst->uc_mcontext.mc_rsp + 512);

    /// align the new stack
    frame = (char*)((((size_t)frame + 256) & -256) | 
                    (dst->uc_mcontext.mc_rsp & 0xff));

    /// copy stack frames
    len = (size_t)exectx->stk_frame - dst->uc_mcontext.mc_rsp;

    memcpy(frame, (char*)dst->uc_mcontext.mc_rsp, len);

    /// source base pointer
    bp = exectx->stk_frame;
    /// source stack pointer
    sp = dst->uc_mcontext.mc_rsp;
    /// the distance between the destination and source stacks
    diff = (size_t)frame - sp;

    FIX_REG(dst->uc_mcontext.mc_rbp);
    FIX_REG(dst->uc_mcontext.mc_rdi);
    FIX_REG(dst->uc_mcontext.mc_rsi);

    FIX_REG(dst->uc_mcontext.mc_rbx);
    FIX_REG(dst->uc_mcontext.mc_rcx);
    FIX_REG(dst->uc_mcontext.mc_rdx);

    FIX_REG(dst->uc_mcontext.mc_r8);
    FIX_REG(dst->uc_mcontext.mc_r9);
    FIX_REG(dst->uc_mcontext.mc_r10);
    FIX_REG(dst->uc_mcontext.mc_r11);
    FIX_REG(dst->uc_mcontext.mc_r12);
    FIX_REG(dst->uc_mcontext.mc_r13);
    FIX_REG(dst->uc_mcontext.mc_r14);
    FIX_REG(dst->uc_mcontext.mc_r15);

    dst->uc_mcontext.mc_rsp = (size_t)frame;

    #else
    #error Wrong architecture!
    #endif

    #undef FIX_REG

    /// switch contexts
    setcontext(dst);
}

/// ---------------------------------------------------------------------------
/// Serial Operator Support Functions
/// ---------------------------------------------------------------------------

static void cpx_intrinsic_lock_queue(cpx_datatype_sync_queue* queue)
{
    while(__sync_val_compare_and_swap(&queue->lock, 0, 1));

}

static void cpx_intrinsic_unlock_queue(cpx_datatype_sync_queue* queue)
{
    queue->lock = 0;
}

static int cpx_intrinsic_hash(size_t addr)
{
    addr = (addr & 0xffffffff) + (size_t)(((unsigned long long)addr) >> 32);

    return (int)(((addr * 0x7f) >> 8) & 0xffff);
}

void cpx_exported_sync(size_t addr)
{
    /// lock queue
    /// increment lock count
    /// if lock count == 1, unlock queue, return
    /// else queue current thread/address, unlock queue, wait while queued, 
    ///      unlock queue, return

    cpx_datatype_sync_queue* queue;
    cpx_datatype_tls_context* curr;
    cpx_datatype_tls_context* tlsctx = cpx_intrinsic_get_tlsctx();

    if(!tlsctx)
    {
        return;
    }
    /// map address to queue
    queue = &cpx_global_sync_queue[cpx_intrinsic_hash(addr)];
    /// lock queue
    cpx_intrinsic_lock_queue(queue);
    /// increment lock count
    queue->lcnt++;
    /// if lock count == 1, unlock queue, return
    if(queue->lcnt == 1)
    {
        queue->win_count = 1;
        queue->winner = tlsctx;
        cpx_intrinsic_unlock_queue(queue);
        return;
    }
    else
    {
        if(queue->winner == tlsctx)
        {
            queue->lcnt--;
            queue->win_count++;
            cpx_intrinsic_unlock_queue(queue);
            return;
        }
    }
    /// else queue current thread/address, unlock queue, wait while queued, 
    ///      unlock queue, return
    if(!queue->head)
    {
        /// queue as the first element in the circular queue
        curr = tlsctx;
        curr->next = curr->prev = curr;
        queue->head = curr;
    }
    else
    {
        /// queue as the last element in the circular queue
        curr = queue->head;
        tlsctx->next = curr;
        tlsctx->prev = curr->prev;
        curr->prev->next = tlsctx;
        curr->prev = tlsctx;
    }
    cpx_intrinsic_unlock_queue(queue);
    /// false wake safe
    while(tlsctx->next)
    {
        wait();
    }
    cpx_intrinsic_unlock_queue(queue);
}

void cpx_exported_unsync(size_t addr)
{
    /// lock queue
    /// decrement lock count
    /// if lock count == 0, unlock queue, wake next, return
    /// else unqueue next, return

    cpx_datatype_tls_context* curr;
    cpx_datatype_sync_queue* queue;
    /// map address to queue
    queue = &cpx_global_sync_queue[cpx_intrinsic_hash(addr)];
    /// lock queue
    cpx_intrinsic_lock_queue(queue);
    /// decrement win count
    if(--queue->win_count)
    {
        /// still the winner, return
        cpx_intrinsic_unlock_queue(queue);
        return;
    }
    /// decrement lock count
    queue->lcnt--;
    /// no winner yet
    queue->winner = 0;
    /// if lock count == 0, unlock queue, return
    if(!queue->lcnt)
    {
        queue->head = 0;
        cpx_intrinsic_unlock_queue(queue);
        return;
    }
    /// else unqueue next, wake next return
    curr = queue->head;
    queue->winner = curr;
    queue->win_count = 1;

    if(queue->lcnt > 1)
    {
        curr->prev->next = curr->next;
        curr->next->prev = curr->prev;
        queue->head = (cpx_datatype_tls_context*)curr->next;
    }
    else
    {
        queue->head = 0;
    }
    curr->prev = curr->next = 0;

    wake(curr);
}

int cpx_exported_trysync(size_t addr)
{
    /// lock queue
    /// if lock count == 0, increment lock count, unlock queue, return 1
    /// else unlock queue, return 0

    int res = 0;
    /// map address to queue
    cpx_datatype_sync_queue* queue = 
        &cpx_global_sync_queue[cpx_intrinsic_hash(addr)];

    cpx_datatype_tls_context* tlsctx = cpx_intrinsic_get_tlsctx();

    if(!tlsctx)
    {
        return 0;
    }
    cpx_intrinsic_lock_queue(queue);

    if(!queue->lcnt)
    {
        queue->lcnt++;
        queue->win_count = 1;
        queue->winner = tlsctx;
        res = 1;
    }
    cpx_intrinsic_unlock_queue(queue);

    return res;
}

void cpx_exported_sync_default()
{
    cpx_datatype_region_context* rgnctx;
    cpx_datatype_tls_context* tlsctx = cpx_intrinsic_get_tlsctx();

    if(tlsctx)
    {
        if((rgnctx = tlsctx->rgnctx))
        {
            pthread_mutex_lock(&rgnctx->head->mutex);
        }
    }
}

void cpx_exported_unsync_default()
{
    cpx_datatype_region_context* rgnctx;
    cpx_datatype_tls_context* tlsctx = cpx_intrinsic_get_tlsctx();

    if(tlsctx)
    {
        if((rgnctx = tlsctx->rgnctx))
        {
            pthread_mutex_unlock(&rgnctx->head->mutex);
        }
    }
}

/// ---------------------------------------------------------------------------
/// Parallel Task Identification Support Functions
/// ---------------------------------------------------------------------------

/// returns 0 in case of error
void* cpx_exported_pid()
{
    cpx_datatype_region_context* rgnctx;
    cpx_datatype_tls_context* tlsctx = cpx_intrinsic_get_tlsctx();

    if(tlsctx)
    {
        if((rgnctx = tlsctx->rgnctx))
        {
            if(rgnctx->type == cpx_const_task)
            {
                rgnctx->type = cpx_const_thread;
                /// decrement the task count
                (void)__sync_sub_and_fetch(&cpx_global_task_no, 1);
                /// release the free thread semaphore
                sem_post(&cpx_global_free_semaphore);
                /// request for one more thread
                if(__sync_sub_and_fetch(&cpx_global_free_thread_no, 1) < 1)
                {
                    if(cpx_intrinsic_create_thread())
                    {
                        return 0;
                    }
                }
            }
        }
        return tlsctx;
    }
    return 0;
}

int cpx_exported_pix()
{
    cpx_datatype_region_context* rgnctx;
    cpx_datatype_tls_context* tlsctx = cpx_intrinsic_get_tlsctx();

    if(tlsctx)
    {
        if((rgnctx = tlsctx->rgnctx))
        {
            return rgnctx->pix;
        }
    }
    return 0;
}

int cpx_exported_pno()
{
    cpx_datatype_region_context* rgnctx;
    cpx_datatype_tls_context* tlsctx = cpx_intrinsic_get_tlsctx();

    if(tlsctx)
    {
        if((rgnctx = tlsctx->rgnctx))
        {
            return rgnctx->pno;
        }
    }
    return 0;
}

/// ---------------------------------------------------------------------------
/// Parallel Task Synchronization Support Functions
/// ---------------------------------------------------------------------------

void cpx_exported_wake(void* pid)
{
    cpx_datatype_tls_context* tlsctx = (cpx_datatype_tls_context*)pid;

    if(tlsctx)
    {
        sem_post(&tlsctx->wakesig);
    }
}

int cpx_exported_wait(int argc, /*int milliseconds*/...)
{
    int ret = 0;
    struct timespec ts;
    long long delay;
    cpx_datatype_tls_context* tlsctx = cpx_intrinsic_get_tlsctx();
    va_list argptr;
    va_start(argptr, argc);

    if(tlsctx)
    {
        if(!argc)
        {
            sem_wait(&tlsctx->wakesig);
        }
        else
        {
            clock_gettime(CLOCK_REALTIME, &ts);

            delay = ts.tv_nsec + ((long long)va_arg(argptr, int)) * 1000000;

            ts.tv_nsec = delay % 1000000000;
            ts.tv_sec += delay / 1000000000;

            while((ret = sem_timedwait(&tlsctx->wakesig, &ts)) == -1 &&
                  errno == EINTR);
        }
    }
    return ret ? 1 : 0;
}

/// ---------------------------------------------------------------------------
/// Data Decomposition Support Functions
/// ---------------------------------------------------------------------------

int cpx_exported_range_start(int size)
{
    cpx_datatype_region_context* rgnctx;
    cpx_datatype_tls_context* tlsctx = cpx_intrinsic_get_tlsctx();

    if(tlsctx)
    {
        if((rgnctx = tlsctx->rgnctx))
        {
            return rgnctx->pix ? (size / rgnctx->pno) * rgnctx->pix + 
                                 (size % rgnctx->pno) : 0;
        }
    }
    return 0;
}

int cpx_exported_range_end(int size)
{
    int pix;
    int size_per_task;
    int size_mod_task;
    cpx_datatype_region_context* rgnctx;
    cpx_datatype_tls_context* tlsctx = cpx_intrinsic_get_tlsctx();

    if(tlsctx)
    {
        if((rgnctx = tlsctx->rgnctx))
        {
            pix = rgnctx->pix;
            size_per_task = size / rgnctx->pno;
            size_mod_task = size % rgnctx->pno;

            if(pix)
            {
                return size_per_task * pix + size_mod_task + size_per_task;
            }
            else
            {
                return size_per_task + size_mod_task;
            }
        }
    }
    return 0;
}

/// ---------------------------------------------------------------------------
/// End of File
/// ---------------------------------------------------------------------------
